'''
Enumerate.py
'''
import telnetlib
import dns.resolver
from optparse import OptionParser
RECURSIVE_SERVERS = {'BN':'61.6.245.174',
                     'DE':'62.26.190.109',
                     'FR':'62.160.101.135',
                     'HK':'59.148.196.254',
                     'IL':'62.90.100.172',
                     'JP':'61.213.248.3',
                     'NG':'41.211.230.173',
                     'PH':'58.69.139.207',
                     'US':'24.173.220.118',
                     }
PUBLIC_ROUTE_SERVERS=['24.222.75.140','67.17.81.187','12.129.193.235', '66.162.47.58', '203.202.125.6', '196.4.160.227']
parser = OptionParser()
parser.add_option("--host", dest="host", help="Hostname to lookup")
parser.add_option("-i", "--iterations", dest="iterations", type="int", default=1, help="Number of times to run tests")
parser.add_option("--debug", action="store_true", default=0, dest="debug", help="Enable debugging messages")
(options, args) = parser.parse_args()

class AuthResolver(dns.resolver.Resolver):

    def __init__(self, resolver_fqdn):
        self.reset()
        resolvers = dns.resolver.query(resolver_fqdn, 'A')
        for resolver in resolvers:
            self.nameservers.append(resolver.to_text())

class RecursiveResolver(dns.resolver.Resolver):
    def __init__(self, ip):
        self.reset()
        self.nameservers.append(ip)

def check_bgp(ip, looking_glass):
    #print 'Checking ip %s from looking glass: %s' % (ip, looking_glass)
    neighbors = []
    t = telnetlib.Telnet(looking_glass)
    read_more = True
    scrp = t.read_until('>')
    data = ''
    t.write('show ip bgp %s\r\n' %ip)
    while read_more:
        data += t.read_until('>', 2)
        end = data[-9:]
        read_more = False
        # TODO FIX READ MORE SUPPORT 
    t.close()
    for line in data.split('\r\n'):
        if line[0:2] == '  ' and line[0:4] != '    ':
            line=line.strip()
        else:
            continue
        if line == 'Not advertised to any peer':
            continue
        try:
            ldata = line.split(',')
            asnpath = ldata[0]
            #print "ASNPath:",asnpath
            asns = asnpath.split()
            neighbor = asns[-2]
            if neighbor == asns[-1]:
                continue
            if neighbor not in neighbors:
                neighbors.append(neighbor)
        except:
            pass
    #print "found neighbors:", neighbors
    return neighbors

def enumerate(qname,indent=""):
    results = {}
    if options.debug:
        print indent + "enumerating Host:", qname
    # Find the authoritative resolver
    split_qname = qname.split('.')
    for i in range(len(split_qname)):
        try:
            tqname = '.'.join(split_qname[i:])
            if options.debug:
                print indent + "dig", tqname, 'NS'
            answers = dns.resolver.query(tqname, 'NS')
            break
        except dns.resolver.NoAnswer:
            continue
        except dns.resolver.NXDOMAIN:
            continue
    # Iterate over the nameservers querying for results
    cnames = []
    for answer in answers:
        if options.debug:
            print indent + "found authresolver", answer.target.to_text()
        aresolver = AuthResolver(answer.target.to_text())
        # Check the CNAME
        try:
            if options.debug:
                print indent + "dig", qname, 'CNAME @%s' % answer.target
            rrset = aresolver.query(qname, 'CNAME', dns.rdataclass.IN, False, None)
            for rr in rrset:
                if rr.target.to_text() not in cnames:
                    cnames.append(rr.target.to_text())
        except dns.resolver.NoAnswer:
            pass
        except Exception as e:
            print e.__class__
            pass
        # Check the A 
        try:
            if options.debug:
                print indent + "dig", qname, 'A @%s' % answer.target
            rrset = aresolver.query(qname, 'A', dns.rdataclass.IN, False, None)
            for rr in rrset:
                ip = rr.to_text()
                if options.debug:
                    print indent + "  " + ip
                if ip not in results.keys():
                    results[ip] = 1
                #else:
                #    results[ip] = results[ip] + 1
        except dns.resolver.NoAnswer:
            pass
        except Exception as e:
            print e.__class__
            pass
    # Enumerate CNAMES
    for cname in cnames:
        sresults = enumerate(cname,indent+"  ")
        for ip in sresults.keys():
            if ip not in results.keys():
                results[ip] = sresults[ip]
            #else:
            #    results[ip] = results[ip] + sresults[ip]
    return results

#print "Checking from current location"
final = {}
for i in range(options.iterations):
    ips = enumerate(options.host)
    for ip in ips.keys():
        if ip not in final.keys():
            final[ip] = ips[ip]
        else:
            final[ip] = final[ip] + ips[ip]
print "Found %d IP Addresses" % len(final.keys())
print final.keys()
for ip in final.keys():
    neighbors = []
    for looking_glass in PUBLIC_ROUTE_SERVERS:
        nlist = check_bgp(ip, looking_glass)
        for neighbor in nlist:
            if neighbor not in neighbors:
                neighbors.append(neighbor)
    if len(neighbors) > 1:
        print "IP: %s is anycasted from at least %d neighboring ASNs: %s" % (ip, len(neighbors), ','.join(neighbors))


#print "Checking from various loctions"
result_sets = [final,]
for country in RECURSIVE_SERVERS.keys():
    rserver = RECURSIVE_SERVERS[country]
    #print "checking",rserver
    rr = RecursiveResolver(rserver)
    rrset = rr.query(options.host, 'A', dns.rdataclass.IN, False, None)
    set = {}
    for rr in rrset:
        ip = rr.to_text()
        if ip not in set:
            set[ip] = 1
        else:
            set[ip] = set[ip] +1
    if set not in result_sets:
        result_sets.append(set)
if len(result_sets) > 1:
    print "This name is being launched using directional DNS (%d Networks found)" % len(result_sets) 
    for result in result_sets:
        print "Network:", result.keys()
